import textwrap

import pytest
from httpx import AsyncClient
from httpx_sse import aconnect_sse

from progtc.types import MessageType


@pytest.mark.asyncio
async def test_execute_code_basic(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        async def main():
            print("main function called")
        await main()
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": [], "code": code},
    ) as event_source:
        async for event in event_source.aiter_sse():
            messages.append(event.json())
    assert messages[0]["type"] == MessageType.SUCCESS
    assert messages[0]["stdout"] == "main function called\n"


@pytest.mark.asyncio
async def test_execute_code_with_func_call(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        async def say_hello():
            print("say_hello function called")
        async def main():
            print("main function called")
            await say_hello()
        await main()
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": [], "code": code},
    ) as event_source:
        async for event in event_source.aiter_sse():
            messages.append(event.json())
    assert messages[0]["type"] == MessageType.SUCCESS
    assert messages[0]["stdout"] == "main function called\nsay_hello function called\n"


@pytest.mark.asyncio
async def test_execute_code_with_tool_call(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        from tools import get_weather
        async def main():
            print("main function called")
            return await get_weather()
        print(await main())
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    ) as event_source:
        async for event in event_source.aiter_sse():
            message = event.json()
            messages.append(message)
            if (
                message["type"] == MessageType.TOOL_CALL
                and message["tool_name"] == "get_weather"
            ):
                await test_client.post(
                    "/tool-results",
                    json={
                        "execution_id": message["execution_id"],
                        "tool_call_id": message["id"],
                        "result": "Sunny",
                    },
                )
    assert messages[0]["type"] == MessageType.TOOL_CALL, messages[0]
    assert messages[0]["tool_name"] == "get_weather"
    assert messages[0]["args"] == []
    assert messages[0]["kwargs"] == {}
    assert messages[1]["type"] == MessageType.SUCCESS
    assert messages[1]["stdout"] == "main function called\nSunny\n"


@pytest.mark.asyncio
async def test_execute_code_with_tool_call_with_args(test_client: AsyncClient) -> None:
    code = textwrap.dedent(
        """\
        from tools import get_weather
        async def main():
            from tools import get_weather
            print("main function called")
            return await get_weather("UK", city="London")
        print(await main())
        """,
    )
    messages = []
    async with aconnect_sse(
        test_client,
        "POST",
        "/execute-code",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    ) as event_source:
        async for event in event_source.aiter_sse():
            message = event.json()
            messages.append(message)
            if (
                message["type"] == MessageType.TOOL_CALL
                and message["tool_name"] == "get_weather"
            ):
                await test_client.post(
                    "/tool-results",
                    json={
                        "execution_id": message["execution_id"],
                        "tool_call_id": message["id"],
                        "result": "Sunny",
                    },
                )
    assert messages[0]["type"] == MessageType.TOOL_CALL, messages[0]
    assert messages[0]["tool_name"] == "get_weather"
    assert messages[0]["args"] == ["UK"]
    assert messages[0]["kwargs"] == {"city": "London"}
    assert messages[1]["type"] == MessageType.SUCCESS
    assert messages[1]["stdout"] == "main function called\nSunny\n"


@pytest.mark.asyncio
async def test_execute_code_step_basic(test_client: AsyncClient) -> None:
    """Code with no tool calls returns success directly."""
    code = textwrap.dedent("""\
        async def main():
            print("hello")
        await main()
    """)
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": [], "code": code},
        timeout=10,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["type"] == MessageType.SUCCESS
    assert data["stdout"] == "hello\n"


@pytest.mark.asyncio
async def test_execute_code_step_syntax_error(test_client: AsyncClient) -> None:
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": [], "code": "def ("},
        timeout=10,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["type"] == MessageType.SYNTAX_ERROR


@pytest.mark.asyncio
async def test_execute_code_step_runtime_error(test_client: AsyncClient) -> None:
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": [], "code": "raise ValueError('boom')"},
        timeout=10,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["type"] == MessageType.RUNTIME_ERROR


@pytest.mark.asyncio
async def test_execute_code_step_tool_call(test_client: AsyncClient) -> None:
    """First call returns tool calls; second call with tool_results returns success."""
    code = textwrap.dedent("""\
        from tools import get_weather
        result = await get_weather("London")
        print(result)
    """)

    # First request — should return tool calls
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    assert len(data["tool_calls"]) == 1
    assert data["tool_calls"][0]["tool_name"] == "get_weather"
    assert data["tool_calls"][0]["args"] == ["London"]

    state = data["state"]

    # Second request — send tool results, server appends to state
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code,
            "state": state,
            "tool_results": ["Sunny"],
        },
        timeout=10,
    )
    assert response.status_code == 200
    data = response.json()
    assert data["type"] == MessageType.SUCCESS
    assert data["stdout"] == "Sunny\n"


@pytest.mark.asyncio
@pytest.mark.parametrize("iteration", range(10))
async def test_execute_code_step_gather(
    test_client: AsyncClient, iteration: int
) -> None:
    """Concurrent tool calls via asyncio.gather return as a batch
    and replay deterministically across repeated runs."""
    cities = [f"city_{i}" for i in range(20)]
    gather_args = ", ".join(f'get_weather("{c}")' for c in cities)
    code = textwrap.dedent(f"""\
        import asyncio
        from tools import get_weather
        results = await asyncio.gather({gather_args})
        print(",".join(results))
    """)
    results = [f"weather_{i}" for i in range(20)]

    # First request — should return all tool calls as a batch
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    assert len(data["tool_calls"]) == 20
    for i, tc in enumerate(data["tool_calls"]):
        assert tc["args"] == [cities[i]], f"tool call {i} has wrong args"

    state = data["state"]

    # Second request — send all results, expect success
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code,
            "state": state,
            "tool_results": results,
        },
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.SUCCESS
    assert data["stdout"] == ",".join(results) + "\n"


@pytest.mark.asyncio
async def test_execute_code_step_multiple_tool_calls(
    test_client: AsyncClient,
) -> None:
    """Sequential tool calls: each round-trip sends tool_results back."""
    code = textwrap.dedent("""\
        from tools import get_weather
        a = await get_weather("London")
        b = await get_weather("Paris")
        print(f"{a},{b}")
    """)

    # First request — returns the first tool call
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    assert data["tool_calls"][0]["args"] == ["London"]

    state = data["state"]

    # Second request — send first result, get second tool call
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code,
            "state": state,
            "tool_results": ["Sunny"],
        },
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    assert data["tool_calls"][0]["args"] == ["Paris"]

    state = data["state"]

    # Third request — send second result, get final success
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code,
            "state": state,
            "tool_results": ["Rainy"],
        },
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.SUCCESS
    assert data["stdout"] == "Sunny,Rainy\n"


@pytest.mark.asyncio
async def test_execute_code_step_tool_call_with_kwargs(
    test_client: AsyncClient,
) -> None:
    code = textwrap.dedent("""\
        from tools import get_weather
        result = await get_weather("UK", city="London")
        print(result)
    """)

    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": ["get_weather"], "code": code},
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    assert data["tool_calls"][0]["args"] == ["UK"]
    assert data["tool_calls"][0]["kwargs"] == {"city": "London"}

    state = data["state"]

    # Replay with result
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code,
            "state": state,
            "tool_results": ["Sunny"],
        },
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.SUCCESS
    assert data["stdout"] == "Sunny\n"


@pytest.mark.asyncio
async def test_execute_code_step_determinism_error(
    test_client: AsyncClient,
) -> None:
    """Replaying with different code produces a determinism error."""
    code_v1 = textwrap.dedent("""\
        from tools import get_weather
        result = await get_weather("London")
        print(result)
    """)
    code_v2 = textwrap.dedent("""\
        from tools import get_weather
        result = await get_weather("Paris")
        print(result)
    """)

    # First request with code_v1
    response = await test_client.post(
        "/execute-code-step",
        json={"tool_names": ["get_weather"], "code": code_v1},
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.TOOL_CALL
    state = data["state"]

    # Second request with code_v2 — tool call args differ from cached hash
    response = await test_client.post(
        "/execute-code-step",
        json={
            "tool_names": ["get_weather"],
            "code": code_v2,
            "state": state,
            "tool_results": ["Sunny"],
        },
        timeout=10,
    )
    data = response.json()
    assert data["type"] == MessageType.DETERMINISM_ERROR



from __future__ import annotations

import ast
import asyncio
import multiprocessing as mp
import queue
import sys
import traceback
from collections.abc import Generator
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from functools import partial
from io import StringIO
from multiprocessing.queues import Queue as MPQueueType
from types import ModuleType
from uuid import uuid4

from progtc.server.tool_call_manager import DeterminismError, ToolCallManager
from progtc.types import (
    CodeDeterminismError,
    CodeRuntimeError,
    CodeSyntaxError,
    ExecuteCodeError,
    ExecuteCodeStepState,
    ExecuteCodeSuccess,
    ToolCall,
    ToolCallResult,
)

CodeExecutionResult = ExecuteCodeSuccess | ExecuteCodeError | CodeDeterminismError

_mp_context = mp.get_context("spawn")


@contextmanager
def _capture_stdout_stderr() -> Generator[tuple[StringIO, StringIO], None, None]:
    stdout = StringIO()
    stderr = StringIO()
    with redirect_stdout(stdout), redirect_stderr(stderr):
        yield stdout, stderr


async def execute_code(
    execution_id: str,
    code: str,
    tool_names: list[str],
    tool_call_queue: MPQueueType[ToolCall],
    tool_call_results_queue: MPQueueType[ToolCallResult],
    code_execution_result_queue: MPQueueType[CodeExecutionResult],
    state: ExecuteCodeStepState | None = None,
) -> None:
    async with ToolCallManager(
        execution_id=execution_id,
        tool_call_queue=tool_call_queue,
        tool_call_results_queue=tool_call_results_queue,
        state=state,
    ) as tool_call_manager:
        # Create a tools module which the code can import tools from
        # and make it availabl in the sys.modules dictionary.
        tools_module = ModuleType("tools")
        for tool_name in tool_names:
            setattr(
                tools_module,
                tool_name,
                partial(
                    tool_call_manager.call_tool,
                    tool_name,
                ),
            )
        sys.modules["tools"] = tools_module

        # Run the code while capturing stdout and stderr.
        with _capture_stdout_stderr() as (stdout, stderr):
            try:
                compiled = compile(
                    code,
                    "<string>",
                    "exec",
                    flags=ast.PyCF_ALLOW_TOP_LEVEL_AWAIT,
                )
            except (SyntaxError, ValueError):
                traceback.print_exc()
                code_execution_result_queue.put(
                    CodeSyntaxError(
                        stdout=stdout.getvalue(),
                        stderr=stderr.getvalue(),
                    ),
                )
                return

            try:
                exec_globals = {"__builtins__": __builtins__}
                coro = eval(compiled, exec_globals)
                if asyncio.iscoroutine(coro):
                    await coro
            except DeterminismError:
                traceback.print_exc()
                code_execution_result_queue.put(
                    CodeDeterminismError(
                        stdout=stdout.getvalue(),
                        stderr=stderr.getvalue(),
                    ),
                )
                return
            except Exception:
                traceback.print_exc()
                code_execution_result_queue.put(
                    CodeRuntimeError(
                        stdout=stdout.getvalue(),
                        stderr=stderr.getvalue(),
                    ),
                )
                return

    code_execution_result_queue.put(
        ExecuteCodeSuccess(stdout=stdout.getvalue(), stderr=stderr.getvalue()),
    )


def execute_code_sync(
    execution_id: str,
    code: str,
    tool_names: list[str],
    tool_call_queue: MPQueueType[ToolCall],
    tool_call_results_queue: MPQueueType[ToolCallResult],
    code_execution_result_queue: MPQueueType[CodeExecutionResult],
    state: ExecuteCodeStepState | None = None,
) -> None:
    try:
        asyncio.run(
            execute_code(
                execution_id=execution_id,
                code=code,
                tool_names=tool_names,
                tool_call_queue=tool_call_queue,
                tool_call_results_queue=tool_call_results_queue,
                code_execution_result_queue=code_execution_result_queue,
                state=state,
            ),
        )
    except Exception as exc:
        code_execution_result_queue.put(
            CodeRuntimeError(
                stdout="",
                stderr=f"Unexpected error in code execution {exc}",
            ),
        )
        return


class CodeExecutionProcess:
    def __init__(
        self,
        code: str,
        tool_names: list[str],
        state: ExecuteCodeStepState | None = None,
    ) -> None:
        self.execution_id = uuid4().hex
        self._tool_call_queue: MPQueueType[ToolCall] = _mp_context.Queue()
        self._tool_call_results_queue: MPQueueType[ToolCallResult] = _mp_context.Queue()
        self._code_execution_result_queue: MPQueueType[CodeExecutionResult] = (
            _mp_context.Queue()
        )
        self._process = _mp_context.Process(
            target=execute_code_sync,
            kwargs=dict(
                execution_id=self.execution_id,
                code=code,
                tool_names=tool_names,
                tool_call_queue=self._tool_call_queue,
                tool_call_results_queue=self._tool_call_results_queue,
                code_execution_result_queue=self._code_execution_result_queue,
                state=state,
            ),
        )

    def start(self) -> None:
        self._process.start()

    def stop(self) -> None:
        if self._process.is_alive():
            self._process.terminate()
        self._process.join()

    def is_done(self) -> bool:
        return self._process.exitcode is not None

    def result(self) -> CodeExecutionResult:
        if not self.is_done():
            raise RuntimeError("Code execution process not done")
        try:
            return self._code_execution_result_queue.get(timeout=0.1)
        except queue.Empty:
            return CodeRuntimeError(
                stdout="",
                stderr=(
                    "Unexpected error in code execution process, "
                    f"exitcode={self._process.exitcode}"
                ),
            )

    def next_tool_call(self, timeout: float = 0.1) -> ToolCall | None:
        try:
            return self._tool_call_queue.get(timeout=timeout)
        except queue.Empty:
            return None

    def next_tool_calls(self, timeout: float = 0.1) -> list[ToolCall]:
        tool_calls = []
        while True:
            try:
                tool_call = self._tool_call_queue.get(timeout=timeout)
                tool_calls.append(tool_call)
            except queue.Empty:
                return tool_calls

    def send_tool_result(self, tool_call_result: ToolCallResult) -> None:
        self._tool_call_results_queue.put(tool_call_result)



import logging
import secrets
import time
from collections.abc import Generator

from fastapi import APIRouter, Depends, FastAPI, Header, HTTPException, Response
from fastapi.responses import StreamingResponse

from progtc.server.code_execution_manager import CodeExecutionManager
from progtc.server.config import server_config
from progtc.types import (
    CodeTimeoutError,
    ExecuteCodeRequest,
    ExecuteCodeStepMessage,
    ExecuteCodeStepRequest,
    ExecuteCodeStepState,
    ExecuteCodeStepToolCalls,
    ToolCallResult,
)

logger = logging.getLogger(__name__)

code_execution_manager = CodeExecutionManager()


def authenticate(
    x_progtc_api_key: str = Header(),
) -> None:
    if not secrets.compare_digest(x_progtc_api_key, server_config.api_key):
        raise HTTPException(status_code=401, detail="Unauthorized")


app = FastAPI()


@app.get("/ping")
def ping() -> Response:
    return Response(status_code=200)


router = APIRouter(dependencies=[Depends(authenticate)])


@router.post("/tool-results")
def add_tool_result(tool_call_result: ToolCallResult) -> None:
    code_execution_manager.send_tool_result(tool_call_result)


@router.post("/execute-code")
def execute_code(body: ExecuteCodeRequest) -> StreamingResponse:
    def stream() -> Generator[str, None, None]:
        try:
            with code_execution_manager.run(body.code, body.tool_names) as process:
                t_start = time.monotonic()
                t_last_event = time.monotonic()

                # Yield tool calls to the client until the process
                # is done or the timeout is reached
                while (
                    not process.is_done()
                    and time.monotonic() - t_start
                    < server_config.code_execution_timeout
                ):
                    tool_call = process.next_tool_call()
                    if tool_call is not None:
                        yield f"data: {tool_call.model_dump_json()}\n\n"
                        t_last_event = time.monotonic()
                    if time.monotonic() - t_last_event > 15.0:
                        yield ": keep alive\n\n"
                        t_last_event = time.monotonic()

                # Yield the final result to the client.
                if process.is_done():
                    result = process.result()
                else:
                    result = CodeTimeoutError(
                        stdout="",
                        stderr="",
                    )
                yield f"data: {result.model_dump_json()}\n\n"
                logger.debug("stream completed successfully")
        except Exception:
            logger.exception("server error during streaming")
            raise

    return StreamingResponse(stream(), media_type="text/event-stream")


@router.post("/execute-code-step")
def execute_code_step(
    body: ExecuteCodeStepRequest,
) -> ExecuteCodeStepMessage:
    state = (
        ExecuteCodeStepState.decode(body.state)
        if body.state
        else ExecuteCodeStepState()
    )
    if body.tool_results:
        state.tool_results.extend(body.tool_results)
    with code_execution_manager.run(body.code, body.tool_names, state) as process:
        t_start = time.monotonic()

        # Return tool calls to the client until the process
        # is done or the timeout is reached
        while (
            not process.is_done()
            and time.monotonic() - t_start < server_config.code_execution_timeout
        ):
            tool_calls = process.next_tool_calls()
            if tool_calls:
                state.tool_call_hashes.extend(tc.call_hash() for tc in tool_calls)
                return ExecuteCodeStepToolCalls(
                    tool_calls=tool_calls,
                    state=state.encode(),
                )

        # Return the final result.
        if process.is_done():
            result = process.result()
        else:
            result = CodeTimeoutError(
                stdout="",
                stderr="",
            )
        return result


app.include_router(router)



from __future__ import annotations

import asyncio
import queue
from multiprocessing.queues import Queue as MPQueueType
from types import TracebackType

from pydantic import JsonValue

from progtc.server.config import server_config
from progtc.types import (
    ExecuteCodeStepState,
    ToolCall,
    ToolCallResult,
    _tool_call_hash,
)


class DeterminismError(Exception):
    pass


class ToolCallManager:
    def __init__(
        self,
        execution_id: str,
        tool_call_queue: MPQueueType[ToolCall],
        tool_call_results_queue: MPQueueType[ToolCallResult],
        state: ExecuteCodeStepState | None = None,
    ) -> None:
        self._execution_id = execution_id
        self._tool_call_queue = tool_call_queue
        self._tool_call_results_queue = tool_call_results_queue
        self._tool_call_results_split_queue: dict[
            str,
            asyncio.Queue[ToolCallResult],
        ] = {}
        self._handle_results_task: asyncio.Task[None] | None = None
        self._shutdown_event = asyncio.Event()
        self._state = state
        self._tool_call_count = 0
        self._lock = asyncio.Lock()

    async def _handle_results_queue(self) -> None:
        while not self._shutdown_event.is_set():
            try:
                tool_call_result = self._tool_call_results_queue.get_nowait()
                await self._tool_call_results_split_queue[
                    tool_call_result.tool_call_id
                ].put(tool_call_result)
            except queue.Empty:
                await asyncio.sleep(0.1)

    async def __aenter__(self) -> ToolCallManager:
        self._handle_results_task = asyncio.create_task(self._handle_results_queue())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._shutdown_event.set()
        if self._handle_results_task is not None:
            await self._handle_results_task

    async def call_tool(
        self,
        tool_name: str,
        *args: JsonValue,
        **kwargs: JsonValue,
    ) -> JsonValue:
        async with self._lock:
            tool_call = ToolCall(
                execution_id=self._execution_id,
                tool_name=tool_name,
                args=args,
                kwargs=kwargs,
            )

            # If we have already received the result for this tool call, return it.
            if self._state is not None and self._tool_call_count < len(
                self._state.tool_results
            ):
                expected_hash = self._state.tool_call_hashes[self._tool_call_count]
                actual_hash = _tool_call_hash(tool_name, args, kwargs)
                if actual_hash != expected_hash:
                    raise DeterminismError(
                        f"Tool call {self._tool_call_count} changed on "
                        f"replay: expected {tool_call.tool_name}() to "
                        f"match the original call, but the arguments "
                        f"differ. The executed code must be deterministic."
                    )
                result_value = self._state.tool_results[self._tool_call_count]
                self._tool_call_count += 1
                return result_value

            self._tool_call_results_split_queue[tool_call.id] = asyncio.Queue[
                ToolCallResult
            ]()
            self._tool_call_queue.put(tool_call)

        try:
            result = await asyncio.wait_for(
                self._tool_call_results_split_queue[tool_call.id].get(),
                timeout=server_config.tool_call_timeout,
            )
            return result.result
        finally:
            del self._tool_call_results_split_queue[tool_call.id]



import asyncio
import contextlib
import functools
import inspect
import random
import time
from collections.abc import Awaitable, Callable
from logging import getLogger
from typing import ParamSpec, TypeVar

from httpx import AsyncClient, AsyncHTTPTransport
from httpx_sse import ServerSentEvent, aconnect_sse
from pydantic import JsonValue, TypeAdapter

from progtc.types import (
    ExecuteCodeError,
    ExecuteCodeMessage,
    ExecuteCodeStepMessage,
    ExecuteCodeSuccess,
    ExecutionEvent,
    ExecutionEventHandler,
    ToolCall,
    ToolCallCompletedEvent,
    ToolCallReceivedEvent,
    ToolCallSentEvent,
    ToolHandler,
)

ExecuteCodeMessageAdapter = TypeAdapter[ExecuteCodeMessage](ExecuteCodeMessage)
ExecuteCodeStepMessageAdapter = TypeAdapter[ExecuteCodeStepMessage](
    ExecuteCodeStepMessage
)

logger = getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")


def _calculate_backoff(
    attempt: int,
    exp_multiplier: float = 1.0,
    exp_min: float = 0.0,
    exp_max: float = 10.0,
    exp_base: float = 2.0,
    random_min: float = 0.0,
    random_max: float = 1.0,
) -> float:
    exp_delay = exp_multiplier * (exp_base**attempt)
    delay = max(min(exp_delay, exp_max), exp_min)
    delay += random.uniform(random_min, random_max)
    return delay


def _retry(
    stop_after_attempt: int = 3,
    exp_multiplier: float = 1.0,
    exp_min: float = 0.0,
    exp_max: float = 10.0,
    exp_base: float = 2.0,
    random_min: float = 0.0,
    random_max: float = 1.0,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    def decorator(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @functools.wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            for attempt in range(stop_after_attempt):
                try:
                    return await fn(*args, **kwargs)
                except Exception:
                    if attempt == stop_after_attempt - 1:
                        raise
                    delay = _calculate_backoff(
                        attempt,
                        exp_multiplier=exp_multiplier,
                        exp_min=exp_min,
                        exp_max=exp_max,
                        exp_base=exp_base,
                        random_min=random_min,
                        random_max=random_max,
                    )
                    logger.warning(
                        "Error in function %s, retrying in %s seconds",
                        fn.__name__,
                        delay,
                        exc_info=True,
                    )
                    await asyncio.sleep(delay)
            raise RuntimeError("Unreachable")

        return wrapper

    return decorator


DEFAULT_PING_TIMEOUT = 5.0
DEFAULT_WAIT_FOR_SERVER_TIMEOUT = 10.0
DEFAULT_CODE_EXECUTION_TIMEOUT = 30.0
DEFAULT_SEND_TOOL_RESULT_TIMEOUT = 10.0


class AsyncProgtcClient:
    def __init__(
        self,
        base_url: str,
        api_key: str,
        headers: dict[str, str] | None = None,
    ):
        self._base_url = base_url
        self._api_key = api_key
        self._headers = headers or {}

    async def ping(
        self,
        timeout: float | None = DEFAULT_PING_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> bool:
        merged_headers = {**self._headers, **(headers or {})}
        async with AsyncClient(
            base_url=self._base_url, headers=merged_headers, timeout=timeout
        ) as client:
            try:
                response = await client.get("/ping")
                return response.status_code == 200
            except Exception:
                return False

    async def wait_for_server(
        self,
        timeout: float | None = DEFAULT_WAIT_FOR_SERVER_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> None:
        t0 = time.monotonic()
        ping_timeout = (
            min(timeout, DEFAULT_PING_TIMEOUT)
            if timeout is not None
            else DEFAULT_PING_TIMEOUT
        )
        while not await self.ping(timeout=ping_timeout, headers=headers):
            await asyncio.sleep(0.5)
            if timeout is not None and time.monotonic() - t0 > timeout:
                raise TimeoutError(f"Progtc server not ready after {timeout}s")

    async def _handle_event(
        self, event: ExecutionEvent, event_handler: ExecutionEventHandler | None
    ) -> None:
        if event_handler is None:
            return
        elif inspect.iscoroutinefunction(event_handler):
            await event_handler(event)
        else:
            event_handler(event)

    async def execute_code(
        self,
        code: str,
        tools: dict[str, ToolHandler],
        timeout: float | None = DEFAULT_CODE_EXECUTION_TIMEOUT,
        headers: dict[str, str] | None = None,
        event_handler: ExecutionEventHandler | None = None,
        send_tool_result_timeout: float | None = DEFAULT_SEND_TOOL_RESULT_TIMEOUT,
    ) -> ExecuteCodeSuccess | ExecuteCodeError:
        merged_headers = {
            **self._headers,
            **(headers or {}),
            "X-Progtc-API-Key": self._api_key,
        }
        async with (
            AsyncHTTPTransport(retries=2) as transport,
            AsyncClient(
                base_url=self._base_url,
                headers=merged_headers,
                timeout=timeout,
                transport=transport,
            ) as client,
            aconnect_sse(
                client,
                "POST",
                "/execute-code",
                json={
                    "tool_names": list(tools.keys()),
                    "code": code,
                },
            ) as event_source,
        ):

            async def handle_tool_call(tool_call: ToolCall) -> None:
                await self._handle_event(
                    ToolCallReceivedEvent(tool_call=tool_call), event_handler
                )
                handler = tools[tool_call.tool_name]
                result = await handler(*tool_call.args, **tool_call.kwargs)
                await self._handle_event(
                    ToolCallCompletedEvent(tool_call=tool_call, result=result),
                    event_handler,
                )

                @_retry()
                async def post_tool_result() -> None:
                    response = await client.post(
                        "/tool-results",
                        json={
                            "execution_id": tool_call.execution_id,
                            "tool_call_id": tool_call.id,
                            "result": result,
                        },
                        timeout=send_tool_result_timeout,
                    )
                    response.raise_for_status()

                await post_tool_result()
                await self._handle_event(
                    ToolCallSentEvent(tool_call=tool_call, result=result), event_handler
                )

            pending_tasks: set[asyncio.Task[None]] = set()
            event_iter = event_source.aiter_sse()
            event_task: asyncio.Task[ServerSentEvent] | None = None

            while True:
                if event_task is None:
                    event_task = asyncio.create_task(anext(event_iter))

                all_tasks: set[asyncio.Task[ServerSentEvent] | asyncio.Task[None]] = {
                    event_task,
                    *pending_tasks,
                }

                done, _ = await asyncio.wait(
                    all_tasks, return_when=asyncio.FIRST_COMPLETED
                )

                # Check for failed tool call tasks
                done_tool_tasks = done & pending_tasks
                pending_tasks -= done_tool_tasks
                for task in done_tool_tasks:
                    if (exc := task.exception()) is not None:
                        event_task.cancel()
                        with contextlib.suppress(asyncio.CancelledError):
                            await event_task
                        raise exc

                # Process event if received
                if event_task in done:
                    try:
                        event = event_task.result()
                    except StopAsyncIteration as err:
                        raise RuntimeError("Stream ended without result") from err
                    finally:
                        event_task = None

                    event_obj = ExecuteCodeMessageAdapter.validate_python(event.json())
                    if isinstance(event_obj, ToolCall):
                        task = asyncio.create_task(handle_tool_call(event_obj))
                        pending_tasks.add(task)
                    else:
                        if pending_tasks:
                            await asyncio.gather(*pending_tasks)
                        return event_obj

    async def execute_code_step(
        self,
        code: str,
        tool_names: list[str],
        state: str | None = None,
        tool_results: list[JsonValue] | None = None,
        timeout: float | None = DEFAULT_CODE_EXECUTION_TIMEOUT,
        headers: dict[str, str] | None = None,
    ) -> ExecuteCodeStepMessage:
        merged_headers = {
            **self._headers,
            **(headers or {}),
            "X-Progtc-API-Key": self._api_key,
        }
        body = {
            "tool_names": tool_names,
            "code": code,
            "state": state,
            "tool_results": tool_results,
        }
        async with AsyncClient(
            base_url=self._base_url,
            headers=merged_headers,
            timeout=timeout,
        ) as client:
            response = await client.post(
                "/execute-code-step",
                json=body,
            )
            response.raise_for_status()
            return ExecuteCodeStepMessageAdapter.validate_python(response.json())



import base64
import gzip
import hashlib
import json
from collections.abc import Awaitable, Callable
from enum import Enum
from typing import Literal, Protocol, TypeAlias
from uuid import uuid4

from pydantic import BaseModel, Field, JsonValue


class ExecuteCodeRequest(BaseModel):
    tool_names: list[str]
    code: str


class MessageType(str, Enum):
    TOOL_CALL = "tool_call"
    SUCCESS = "success"
    SYNTAX_ERROR = "syntax_error"
    RUNTIME_ERROR = "runtime_error"
    TIMEOUT_ERROR = "timeout_error"
    DETERMINISM_ERROR = "determinism_error"


def _tool_call_hash(
    tool_name: str,
    args: tuple[JsonValue, ...],
    kwargs: dict[str, JsonValue],
) -> str:
    data = json.dumps(
        {"tool_name": tool_name, "args": args, "kwargs": kwargs},
        sort_keys=True,
    )
    return hashlib.sha256(data.encode()).hexdigest()[:16]


class ToolCall(BaseModel):
    type: Literal[MessageType.TOOL_CALL] = MessageType.TOOL_CALL
    id: str = Field(default_factory=lambda: uuid4().hex)
    execution_id: str
    tool_name: str
    args: tuple[JsonValue, ...]
    kwargs: dict[str, JsonValue]

    def call_hash(self) -> str:
        return _tool_call_hash(self.tool_name, self.args, self.kwargs)


class ExecuteCodeSuccess(BaseModel):
    type: Literal[MessageType.SUCCESS] = MessageType.SUCCESS
    stdout: str
    stderr: str


class CodeSyntaxError(BaseModel):
    type: Literal[MessageType.SYNTAX_ERROR] = MessageType.SYNTAX_ERROR
    stdout: str
    stderr: str


class CodeRuntimeError(BaseModel):
    type: Literal[MessageType.RUNTIME_ERROR] = MessageType.RUNTIME_ERROR
    stdout: str
    stderr: str


class CodeTimeoutError(BaseModel):
    type: Literal[MessageType.TIMEOUT_ERROR] = MessageType.TIMEOUT_ERROR
    stdout: str
    stderr: str


class CodeDeterminismError(BaseModel):
    type: Literal[MessageType.DETERMINISM_ERROR] = MessageType.DETERMINISM_ERROR
    stdout: str
    stderr: str


ExecuteCodeError: TypeAlias = CodeSyntaxError | CodeRuntimeError | CodeTimeoutError


ExecuteCodeMessage: TypeAlias = ToolCall | ExecuteCodeSuccess | ExecuteCodeError


class ToolCallResult(BaseModel):
    execution_id: str
    tool_call_id: str
    result: JsonValue


ToolHandler: TypeAlias = Callable[..., Awaitable[JsonValue]]


class ToolCallReceivedEvent(BaseModel):
    type: Literal["tool_call_received"] = "tool_call_received"
    tool_call: ToolCall


class ToolCallCompletedEvent(BaseModel):
    type: Literal["tool_call_completed"] = "tool_call_completed"
    tool_call: ToolCall
    result: JsonValue


class ToolCallSentEvent(BaseModel):
    type: Literal["tool_call_sent"] = "tool_call_sent"
    tool_call: ToolCall
    result: JsonValue


ExecutionEvent: TypeAlias = (
    ToolCallReceivedEvent | ToolCallCompletedEvent | ToolCallSentEvent
)


class SyncExecutionEventHandler(Protocol):
    def __call__(self, event: ExecutionEvent) -> None: ...


class AsyncExecutionEventHandler(Protocol):
    def __call__(self, event: ExecutionEvent) -> Awaitable[None]: ...


ExecutionEventHandler: TypeAlias = (
    SyncExecutionEventHandler | AsyncExecutionEventHandler
)


class ExecuteCodeStepState(BaseModel):
    tool_results: list[JsonValue] = Field(default_factory=list)
    tool_call_hashes: list[str] = Field(default_factory=list)

    def encode(self) -> str:
        data = self.model_dump_json().encode()
        return base64.b64encode(gzip.compress(data)).decode()

    @classmethod
    def decode(cls, encoded: str) -> "ExecuteCodeStepState":
        data = gzip.decompress(base64.b64decode(encoded))
        return cls.model_validate_json(data)


class ExecuteCodeStepRequest(BaseModel):
    tool_names: list[str]
    code: str
    tool_results: list[JsonValue] | None = None
    state: str | None = None


class ExecuteCodeStepToolCalls(BaseModel):
    type: Literal[MessageType.TOOL_CALL] = MessageType.TOOL_CALL
    tool_calls: list[ToolCall]
    state: str


ExecuteCodeStepMessage: TypeAlias = (
    ExecuteCodeStepToolCalls
    | ExecuteCodeSuccess
    | ExecuteCodeError
    | CodeDeterminismError
)
